#include <ostream>
#include <unistd.h>  // getopt
#include "MCFgen.hpp"  // die, tostring
#include "args.hpp"

namespace args {
  string command = "glam";
  string seqfile;
  // set defaults:
  uint runs = 10;
  uint stop_after = 10000;
  bool one_strand = false;
  uint min_width = 1;
  uint max_width = 10000;
  bool zoops = true;
  string old_glam_file;
  bool strict_subopt = false;
  uint flank_seq = 0;
  bool mask_lower = false;
  float resize_freq = 1;
  float init_temp = 0.9;
  float cool_factor = 1;
  bool lam = false;
  bool even_bg = false;
  float pseudocount_weight = 1.5;  // 1.5 is best fit to TRANSFAC matrices
  bool even_pseudos = false;
  uint random_seed = 1;  // if srand() isn't called at all, seed=1
  bool all_aligns = false;
  bool profile = false;
  bool want_evalue = true;
}

void args::parse(int argc, char **argv)
{
  using mcf::tostring;

  const string doc = "\
GLAM is a program for discovering biologically significant motifs shared by a\n\
set of nucleotide sequences. GLAM takes as input a file of nucleotide sequences\n\
in fasta format. Any non-alphabetic characters in the sequences are ignored, and\n\
any alphabetic characters except A, C, G, T (uppercase or lowercase) are\n\
converted to 'n' and excluded from alignment. If the file is called 'myseqs.fa',\n\
run GLAM on it by typing:\n\
\n\
  glam myseqs.fa\n\
\n\
The output will look something like this:\n\
\n\
  GLAM: Gapless Local Alignment of Multiple sequences\n\
  Compiled on Dec 11 2003\n\
\n\
  Run 1... 10345 iterations\n\
  Run 2... 15190 iterations\n\
  Run 3... 24749 iterations\n\
  Run 4... 20294 iterations\n\
  Run 5... 20488 iterations\n\
  Run 6... 16930 iterations\n\
  Run 7... 21583 iterations\n\
  Run 8... 23260 iterations\n\
  Run 9... 13733 iterations\n\
  Run 10... 14805 iterations\n\
  Calculating score distribution...\n\
  Calculating random walk parameters...\n\
\n\
  Best alignment found:\n\
  Score: 44.7738 bits  Width: 32  Sequences: 5  Runs: 6  E-value: 2.29\n\
  FirstSeq     164 GGACTAAGTTACTTAAACTGTTCAGGAGATAC 195  +  (8.36)\n\
  2ndSeq       244 GGGCATGGTGACCTTTCGCACTCTGGGCATGC 275  +  (9.91)\n\
  3rdSeq       244 GGTCAAGGTCACCGACAGCAGTAAGGGCTGAC 275  +  (11.6)\n\
  4thSeq       244 GGGCAAAGTGACTGGACATAGGAGTGGGACAC 275  +  (10.4)\n\
  LastSeq       92 GGGCAAAGCAACATAGCGGGGTAGGGTCCTCC  61  -  (7.88)\n\
                   ** ****** ** *  *  * ** ** *   *\n\
\n\
  Other alignments:\n\
  Score: 43.4623 bits  Width: 38  Sequences: 5  Runs: 1  E-value: 5.68\n\
  Score: 43.2357 bits  Width: 19  Sequences: 5  Runs: 3  E-value: 6.65\n\
\n\
  glam myseqs.fa\n\
  5 sequences in file\n\
  Residue abundances: a=0.248002 c=0.251998 g=0.251998 t=0.248002\n\
  Pseudocounts: a=0.372002 c=0.377998 g=0.377998 t=0.372002\n\
  Max possible alignment width: 500\n\
  K: 0.178944\n\
  H: 1.25899\n\
\n\
GLAM works by starting from a completely random alignment of the sequences, and\n\
making small refinements to it over many iterations, in an attempt to find the\n\
best possible alignment. Since this procedure does not guarantee finding the\n\
optimum alignment, GLAM repeats it 10 times from different starting points\n\
(10 runs). If several runs converge to the same best alignment, we have\n\
increased confidence that it is indeed the optimum alignment. In this example 6\n\
out of 10 runs gave the same best alignment.\n\
\n\
The score is GLAM's measure of how strong/well-conserved/striking the alignment\n\
is: the higher the better. The E-value indicates how often we would expect an\n\
alignment of this score or greater to exist among unrelated sequences just by\n\
chance. We hope to find E-values lower than 1, but in this example it is only\n\
2.29, so the alignment does not appear to be statistically significant. The\n\
stars indicate conserved columns that contribute positively to the score. The\n\
numbers in brackets are the marginal scores (in bits) of each segment in the\n\
alignment: i.e. the score gained by including this segment in the alignment\n\
rather than excluding it. We might feel more confident that segments with higher\n\
marginal scores are true motif instances. The marginal scores won't in general\n\
sum to the total alignment score.\n\
\n\
REPETITIVE ELEMENTS\n\
\n\
One of the chief problems when using GLAM is the presence of ubiquitous\n\
repetitive or \"low complexity\" elements, such as Alus or A-rich tracts. These\n\
elements can often be aligned with extremely high statistical significance,\n\
perhaps overshadowing more interesting motifs. We suggest two ways to deal with\n\
this problem. One is to mask these elements prior to alignment, using programs\n\
such as RepeatMasker, which specializes in interspersed repeats, and nseg and\n\
dust, which specialize in low complexity elements and simple sequence repeats.\n\
The other is to apply GLAM repeatedly using the -g option to uncover secondary\n\
alignments beyond the strongest one.\n\
\n\
OPTIONS\n\
\n\
There are many options for modifying GLAM's behavior. We describe them in\n\
approximate order of importance: many of them are rather specialized and you\n\
don't need to worry about them.\n\
\n\
 -h  Help: print documentation. You already know this one.\n\
 -n  This important parameter controls the tradeoff between speed and accuracy.\n\
     If you don't play around with any other parameters, play around with this\n\
     one. Each alignment run will continue until n (default = 10000) iterations\n\
     have passed without improving on the best alignment found so far. We like\n\
     to set n sufficiently high that at least 3 out of 10 runs converge to the\n\
     same alignment. Low values of n are adequate when the problem size is\n\
     small, i.e. when the sequences are short and more importantly there are few\n\
     of them, but high values of n are needed for large problems. In addition,\n\
     smaller values of n are sufficient when there is a strong alignment to be\n\
     found, but larger values are necessary when there isn't, e.g. for finding\n\
     the optimal alignment of random sequences. You'll have to choose n on a\n\
     case-by-case basis, but to give some examples we have used n=1000 to align\n\
     5 x 500bp sequences, and n=20000 to align 20 x 1000bp sequences. For larger\n\
     problems it may be impossible to converge reproducibly to the same exact\n\
     alignment in reasonable time, but in these cases you can check that similar\n\
     motifs are reproducibly obtained.\n\
 -r  The number of alignment runs (default = 10).\n\
 -1  (The digit, not the letter): just examine the direct strand (default = both\n\
     strands).\n\
 -a  Minimum alignment width (default = 1).\n\
 -b  Maximum alignment width (default = 10000).\n\
 -z  Require every sequence to participate in the alignment.\n\
 -g  Supply a previous GLAM output file, and exclude the best alignment found\n\
     previously from being recovered again. The previous GLAM output will be\n\
     appended to the current output: if a file with multiple such outputs is\n\
     supplied with -g, all best alignments found previously will be excluded.\n\
 -k  Prevent all residues participating in previous alignments from\n\
     participating in this one. The default behavior is that any pair of\n\
     residues aligned previously may not be aligned this time.\n\
 -l  (The letter, not the digit): exclude lowercase letters from being aligned.\n\
     Lowercase letters are often used to indicate repetitive sequence.\n\
 -v  Verbose: if multiple runs return more than one alignment, print all\n\
     alignments in full.\n\
 -f  Print this number of flanking residues, in lowercase, either side of the\n\
     alignment (default = 0).\n\
 -q  Pretend that the background residue abundances equal 1/4, instead of\n\
     estimating them from the input sequences. This option might be a good idea\n\
     for aligning very short sequences that are mostly covered by the motif.\n\
 -d  Frequency of width-adjusting versus sequence-adjusting moves (default = 1).\n\
     When the number of sequences is large compared to the sequence length, GLAM\n\
     has difficulty widening the alignment, and may return excessively narrow\n\
     alignments. Increasing the frequency of width-adjusting moves compensates\n\
     for this problem to some extent. In theory this problem can always be\n\
     overcome by making the -n parameter sufficiently large.\n\
\n\
Temperature options: GLAM's alignment strategy has a concept of 'temperature'.\n\
At low temperatures it strongly favors refinements that improve alignment\n\
quality, and at high temperatures it only weakly favors such refinements. If the\n\
temperature is too low it will get stuck in a local optimum, and if the\n\
temperature is too high it will never find good alignments. The default strategy\n\
(constant t=0.9) seems to work well in a variety of cases.\n\
\n\
 -t  Initial temperature (default = 0.9).\n\
 -c  Cooling factor: multiply temperature by this amount each iteration\n\
     (default = 1, i.e. constant temperature).\n\
 -m  Use the 'modified Lam schedule' instead of the default geometric schedule.\n\
     This is a strategy where the algorithm aims to achieve a target 'accept\n\
     rate', i.e. rate of altering the alignment versus leaving it unchanged per\n\
     iteration. In the early phase of the algorithm, the target accept rate\n\
     decays geometrically from 100% to 44%. In the middle phase it remains\n\
     constant at 44%. In the final phase it decays geometrically to 0%.\n\
     Whenever the actual accept rate is higher than the target the temperature\n\
     is multiplied by c, when it is lower than the target the temperature is\n\
     divided by c. If -m is selected, the -n option indicates the total number\n\
     of iterations per run, not the number of iterations without improvement.\n\
 -w  Print energy (negative alignment score in nats), accept rate, temperature,\n\
     width, and number of sequences in the alignment after each iteration.\n\
 -p  Pseudocount weight (default = 1.5).\n\
 -u  Uniform pseudocounts: set each pseudocount equal to p/4.\n\
     Default = p * (background residue abundances).\n\
 -s  Seed for the random number generator (default = 1).\n\
 -e  Turn off the E-value calculation. If it's too slow for you.\n\
\n\
Example:\n\
  glam -n5000 -l -v -f5 myseqs.fa\n\
\n\
KNOWN LIMITATIONS\n\
\n\
- The E-values become increasingly conservative as the number of sequences\n\
  increases.\n\
- If the number of sequences is many-fold larger than the sequence length, GLAM\n\
  has difficulty widening the alignment.\n\
- The E-value calculation aborts on more than about 730 input sequences.\n\
\n\
PUBLICATION\n\
\n\
Martin C Frith, Ulla Hansen, John L Spouge, Zhiping Weng (in press)\n\
Finding functional sequence elements by multiple local alignment\n\
Nucleic Acids Research\n\
";

  const string usage =
    "Usage summary: glam [options] myseqs.fa\n"
    "Options:\n"
    "-h  help: print documentation\n"
    "-n  end each run after this many iterations without improvement (" + tostring(stop_after) + ")\n"
    "-r  number of alignment runs (" + tostring(runs) + ")\n"
    "-1  (\"one\") just examine forward strand (default = both strands)\n"
    "-a  minimum alignment width (" + tostring(min_width) + ")\n"
    "-b  maximum alignment width (" + tostring(max_width) + ")\n"
    "-z  turn off ZOOPS (force every sequence to participate in the alignment)\n"
    "-g  previous GLAM alignment(s) to exclude\n"
    "-k  hard-exclude previous alignments\n"
    "-l  (\"ell\") filter lowercase letters\n"
    "-v  verbose: print all alignments in full\n"
    "-f  number of flanking residues to print either side of the alignment (" + tostring(flank_seq) + ")\n"
    "-q  pretend residue abundances = 1/4\n"
    "-d  frequency of width-adjusting moves (" + tostring(resize_freq) + ")\n"
    "-t  initial temperature (" + tostring(init_temp) + ")\n"
    "-c  cooling factor (" + tostring(cool_factor) + ")\n"
    "-m  use modified Lam schedule (default = geometric schedule)\n"
    "-w  print progress information after each iteration\n"
    "-p  pseudocount weight (" + tostring(pseudocount_weight) + ")\n"
    "-u  use uniform pseudocounts: each pseudocount = p/4\n"
    "-s  seed for random number generator (" + tostring(random_seed) + ")\n"
    "-e  turn off the E-value calculation\n"
    ;

  int c;

  for (int i = 1; i < argc; ++i)
    command += string(" ") + argv[i];

  while ((c = getopt(argc, argv, "hn:r:1a:b:zg:klvf:qd:t:c:mwp:us:e")) != -1)
    switch (c) {
    case 'h':
      cout << doc << endl;
      exit(0);
    case 'n':
      stop_after = atoi(optarg);
      break;
    case 'r':
      runs = atoi(optarg);
      break;
    case '1':
      one_strand = true;
      break;
    case 'a':
      min_width = atoi(optarg);
      if (min_width < 1) mcf::die("Error: minimum alignment width should be >= 1");
      break;
    case 'b':
      max_width = atoi(optarg);
      break;
    case 'z':
      zoops = false;
      break;
    case 'g':
      old_glam_file = optarg;
      break;
    case 'k':
      strict_subopt = true;
      break;
    case 'l':
      mask_lower = true;
      break;
    case 'v':
      all_aligns = true;
      break;
    case 'f':
      flank_seq = atoi(optarg);
      break;
    case 'q':
      even_bg = true;
      break;
    case 'd':
      resize_freq = atof(optarg);
      if (resize_freq < 0) mcf::die("Error: width adjustment frequency should be >= 0");
      break;
    case 't':
      init_temp = atof(optarg);
      break;
    case 'c':
      cool_factor = atof(optarg);
      break;
    case 'm':
      lam = true;
      break;
    case 'w':
      profile = true;
      break;
    case 'p':
      pseudocount_weight = atof(optarg);
      break;
    case 'u':
      even_pseudos = true;
      break;
    case 's':
      random_seed = atoi(optarg);
      break;
    case 'e':
      want_evalue = false;
      break;
    case '?':
      mcf::die("\n" + usage);  // "invalid option" message is printed by getopt
    }

  if (max_width < min_width) mcf::die("Error: max width should be >= min width");

  if (optind + 1 != argc)  // there should be 1 more non-option argument
    mcf::die("Input error\n\n" + usage);

  seqfile = argv[optind];
}
