// GLAM: Gapless Local Alignment of Multiple sequences
// Based on Gibbs sampling
// Automatically determines alignment width
// Calculates E-values using a BLAST-like scheme

#include <algorithm>  // sort, max, min, transform
#include <fstream>
#include <functional>  // plus, multiplies, bind2nd
#include <iomanip>  // setw
#include <iostream>
#include <sstream>
#include <cctype>  // toupper
#include <cmath>  // log, exp, lgamma (non-standard??)
#include <cstdlib>  // srand

#include "matrix.hpp"
#include "MCFgen.hpp"
#include "MCFbio.hpp"
#include "args.hpp"
#include "stats.hpp"
#include "glam.hpp"

// how many digits needed to write the number in base 10?
// what's the coolest way to do this?
unsigned num_digits(unsigned x)
{
  unsigned n = 1;
  while (x /= 10) ++n;
  return n;
}

// DNA-to-number translator that only recognizes uppercase
inline unsigned nolower_translator(char c)
{
  switch(c) {
  case 'A':
    return 0u;
  case 'C':
    return 1u;
  case 'G':
    return 2u;
  case 'T':
    return 3u;
  default:
    return 4u;
  }
}

namespace glam {
  using mcf::matrix;
  const float minus_infinity = -1e+25;  // FLT_MAX is at least 1e+37
  const uint alphsize = 4;  // size of alphabet

  // struct to store results of an annealing run:
  struct result_struct {
    alignment aln;  // the alignment
    uint block_num;  // number of sequences containing a block
    float score;
    uint count;  // how many times we get the same result
  };

  vector<vector<uint> > seqs;  // the sequences, represented as integers with ACGTN=01234
  vector<uint> seq_lengths;  // lengths of un-preprocessed sequences
  vector<string> seq_names;
  uint max_poss_width;  // maximum possible width of the alignment

  vector<float> bg_probs;  // background abundances of ACGT
  vector<float> pseudos;  // which Dirichlet prior to use for the motif

  // lookup tables:
  matrix<float> precalc_gibbs_mat;  // used with regular Gibbs moves
  matrix<float> precalc_resize_scores;  // used with resizing moves
  vector<float> precalc_column;  // base score for 1 column, depends on number of blocks present
  vector<float> precalc_absent;  // score for block being absent in one sequence (depends on # of blocks)
  vector<float> width_probs;  // a priori probs of different alignment widths
  vector<alignment> old_alns;  // alignments found previously: avoid overlap

  alignment align;  // the current alignment
  uint block_num;  // # of sequences containing a block: equals # of nonzero entries in align.starts
  vector<vector<uint> > count_matrix;  // holds counts of ACGT in each column

  //  float ** gibbs_mat;  // 177.420 seconds on Libby's data
  float (*gibbs_mat)[alphsize+1];  // 177.830 seconds on Libby's data (using malloc. 177.140 using new)
  //  matrix<float> gibbs_mat;  // 195.760 seconds on Libby's data. Equally bad if [] returns an iterator or a pointer

  float inv_temp;  // = 1 / temperature

  vector<result_struct> results;  // stores results of multiple annealing runs

  // functions:
  void rescale_exp(vector<float> & scores);
  unsigned choose_score(vector<float> & scores);
  void count_seq(vector<vector<uint> > & mat,
		 const vector<uint>::const_iterator & start);
  void uncount_seq(vector<vector<uint> > & mat,
		   const vector<uint>::const_iterator & start);
  bool count_column(vector<uint> & counts, const vector<uint> & starts, int offset);
  float counts_to_score(const vector<uint> & counts);
  void renormalize_scores();
  float get_score(const vector<vector<uint> > & count_mat, uint block_num);
  void construct_gibbs_matrix(const vector<vector<uint> > & counts, uint block_num);
  void subopt_strict(vector<float> & scores, const vector<alignment> & old_alns, const alignment & aln, uint seq);
  void subopt_lenient(vector<float> & scores, const vector<alignment> & old_alns, const alignment & aln, uint seq);
  bool gibbs_move();
  bool resize_right_clash(const vector<alignment> & old_alns, const vector<uint> & starts, uint width);
  bool resize_right();
  bool resize_left_clash(const vector<alignment> & old_alns, const vector<uint> & starts, uint width);
  bool resize_left();
  bool do_move();
  void get_count_matrix(vector<vector<uint> > & count_mat, const alignment & aln);
  bool aln_overlap(const alignment & a1, const alignment & a2);
  bool any_aln_overlap(const alignment & a, const vector<alignment> & v);
  bool overlaps_mask(const vector<uint> & seq, uint start, uint width);
  bool overlaps_alns(const vector<alignment> & alns, uint start, uint width, uint seq);
  void random_aln(alignment & aln);
  void random_start();
  void anneal(result_struct & best);

  // initialization & preprocessing functions:
  void mask_alns(vector<uint> & seq, const vector<alignment> & alns, uint s);
  void add_width_ways(vector<uint> & num_ways, uint run);
  unsigned get_width_probs(vector<float> & probs);
  void preprocess_seqs(bool one_strand);
  void init_gibbs_mat(uint width);
  void init();
  void read_seqs(const string & filename, bool one_strand);

  void print_misc_info(ostream & strm);
  bool canonical_orientation(const alignment & a, const vector<uint> & seqlens);
  void add_to_results(result_struct & res);
  float get_marginal_score(const result_struct & r, uint s);
  void print_alignment(const result_struct & r);
  void print_results(bool one_strand, float k, float h);
}

// Return true if the ranges overlap
// start == 0 indicates there is no range
inline bool ranges_overlap(const unsigned start1, const unsigned width1,
			   const unsigned start2, const unsigned width2) {
  return start1+width1 > start2 && start2+width2 > start1 && start1 && start2;
}

// Exponentiate the scores, rescaling to avoid overflow
void glam::rescale_exp(vector<float> & scores)
{
  assert(!scores.empty());

  const float max = *max_element(scores.begin(), scores.end());
  for (vector<float>::iterator s = scores.begin(); s != scores.end(); ++s)
    *s = exp(*s - max);
}

// Stochastically pick a score, with probability proportional to exp(score)
unsigned glam::choose_score(vector<float> & scores)
{
  rescale_exp(scores);
  // to replicate old results: use temperature parameter -t 1
  // and replace the line below with the 2 lines below that
  const float tot = accumulate(scores.begin(), scores.end(), 0.0f);
  //  static float tot;  tot = 0;
  //  for (uint i = 0; i < scores.size(); ++i)  tot += scores[i];
  return random_choice(scores.begin(), scores.end(), tot) - scores.begin();
}

// Add the residues beginning at "start" to the count matrix
void glam::count_seq(vector<vector<uint> > & mat,
		     const vector<uint>::const_iterator & start)
{
  for (uint k = 0; k < mat.size(); ++k)
    ++mat[k][*(start + k)];
}

// Subtract the residues beginning at "start" from the count matrix
void glam::uncount_seq(vector<vector<uint> > & mat,
		       const vector<uint>::const_iterator & start)
{
  for (uint k = 0; k < mat.size(); ++k)
    --mat[k][*(start + k)];
}

// Count bases in a column offset from start of motif
// return false if there's a masked base
// assumes counts has been suitably initialized
bool glam::count_column(vector<uint> & counts, const vector<uint> & starts, const int offset)
{
  for (uint s = 0; s < starts.size(); ++s)  // count bases in this column
    if (starts[s] != 0) {  // if the block is present in this sequence
      const uint i = seqs[s][starts[s] + offset];  // works if offset < 0
      if (i < alphsize)
	++counts[i];
      else
	return false;
    }

  return true;
}

// Calculate score of a column of the motif from the counts of ACGT
// doesn't include the part of the score that depends on block_num but not on counts of ACGT
float glam::counts_to_score(const vector<uint> & counts)
{
  float score = 0;

  for (uint i = 0; i < alphsize; ++i)
    score += precalc_resize_scores[counts[i]][i];

  return score;
}

// Calculate the score of an alignment, given a count matrix
// block_num is the number of sequences in the alignment
float glam::get_score(const vector<vector<uint> > & count_mat, const uint block_num)
{
  const uint width = count_mat.size();
  float score = width * precalc_column[block_num];

  for (uint k = 0; k < width; ++k)
    score += counts_to_score(count_mat[k]);

  return score;
}

// Construct the PSSM used in Gibbs moves from the count matrix
void glam::construct_gibbs_matrix(const vector<vector<uint> > & counts, const uint block_num)
{
  const float norm = precalc_absent[block_num];

  for (uint k = 0; k < counts.size(); ++k)
    for (uint i = 0; i < alphsize; ++i)
      gibbs_mat[k][i] = inv_temp * (precalc_gibbs_mat[counts[k][i]][i] - norm);
}

// Don't allow any previously aligned nucleotide to be aligned
void glam::subopt_strict(vector<float> & scores, const vector<alignment> & old_alns, const alignment & aln, const uint seq)
{
  for (vector<alignment>::const_iterator a = old_alns.begin();
       a != old_alns.end(); ++a) {
    const uint old_start = a->starts[seq];
    if (old_start == 0)
      continue;
    const uint old_end = old_start + a->width;
    const uint begin = aln.width <= old_start ? old_start + 1 - aln.width : 1;
    const uint end = old_end < scores.size() ? old_end : scores.size();
    for (uint k = begin; k < end; ++k)
      scores[k] = minus_infinity;
  }
}

// Don't allow any previously aligned pair of nucleotides to be aligned
void glam::subopt_lenient(vector<float> & scores, const vector<alignment> & old_alns, const alignment & aln, const uint seq)
{
  for (vector<alignment>::const_iterator a = old_alns.begin();
       a != old_alns.end(); ++a) {
    const uint old_start = a->starts[seq];
    if (old_start == 0)
      continue;
    for (uint s = 0; s != aln.starts.size(); ++s)
      if (ranges_overlap(aln.starts[s], aln.width, a->starts[s], a->width)) {
        const uint x = old_start + aln.starts[s] - a->starts[s];
        if (x > 0 && uint(x) < scores.size())  // x < 0 --> x > scores.size() ?
          scores[x] = minus_infinity;
      }
  }
}

bool glam::gibbs_move()
{
  //  cerr << "gibbs_move" << endl;
  // 1. Randomly choose a sequence:
  const uint chosen_seq = rand((uint)seqs.size());
  const vector<uint> & sq = seqs[chosen_seq];

  // 2. Erase its motif:
  const uint old_pos = align.starts[chosen_seq];
  if (old_pos != 0) {
    uncount_seq(count_matrix, sq.begin() + old_pos);
    --block_num;
  }

  // 3. Get score matrix from count matrix:
  construct_gibbs_matrix(count_matrix, block_num);

  // 4. Get scores for each motif position in the chosen sequence:
  // number of scores to get (last base is masked, no need to check it)
  uint score_num = align.width < sq.size() ? sq.size() - align.width : 1;
  vector<float> scores;
  scores.reserve(score_num);
  scores.push_back(args::zoops ? 0 : minus_infinity);  // score if motif absent

  for (uint n = 1; n < score_num; ++n) {  // the sequence starts from 1
    register float score = 0;
    //    const uint * const p = &sq[n];  // doesn't seem to make it any faster
    for (register uint k = 0; k < align.width; ++k)
      score += gibbs_mat[k][sq[n + k]];
    scores.push_back(score);
  }

  // mask positions forbidden because of previous alignments
  align.starts[chosen_seq] = 0;
  if (args::strict_subopt)
    subopt_strict(scores, old_alns, align, chosen_seq);
  else
    subopt_lenient(scores, old_alns, align, chosen_seq);

  // 5. Stochastically pick a new motif position
  const uint new_pos = choose_score(scores);

  // 6. Add new motif:
  align.starts[chosen_seq] = new_pos;
  if (new_pos != 0) {  // if it's not chosen to be absent
    count_seq(count_matrix, sq.begin() + new_pos);
    ++block_num;
  }

  return new_pos != old_pos;
}

// Would a 1 bp extension clash with a previous alignment?
bool glam::resize_right_clash(const vector<alignment> & old_alns, const vector<uint> & starts, const uint width)
{
  for (vector<alignment>::const_iterator a = old_alns.begin();
       a != old_alns.end(); ++a) {
    uint danger_zone = 0;
    for (uint s = 0; s < starts.size(); ++s)
      if (starts[s] + width == a->starts[s] && starts[s] != 0)
	++danger_zone;
    if (danger_zone > !args::strict_subopt)
      return true;
  }

  return false;
}

// Holds the left end of the motif fixed, and varies the right end
bool glam::resize_right()
{
  //  cerr << "resize_right" << endl;
  const float col_score = precalc_column[block_num];
  vector<float> scores;
  scores.reserve(max_poss_width + 1 - args::min_width);

  count_matrix.resize(args::min_width);
  float score = 0;
  scores.push_back(score);

  // consider block_num == 0
  for (uint k = args::min_width; k < max_poss_width; ++k) {
    if (resize_right_clash(old_alns, align.starts, k))
      break;
    vector<uint> counts(alphsize, 0);
    if (!count_column(counts, align.starts, k))  // stop if hit a masked base
      break;
    count_matrix.push_back(counts);
    score += inv_temp * (col_score + counts_to_score(counts));
    scores.push_back(score);
  }

  const uint old_width = align.width;
  align.width = choose_score(scores) + args::min_width;
  count_matrix.resize(align.width);

  return align.width != old_width;
}

// Would a 1 bp extension clash with a previous alignment?
bool glam::resize_left_clash(const vector<alignment> & old_alns, const vector<uint> & starts, const uint width)
{
  for (vector<alignment>::const_iterator a = old_alns.begin();
       a != old_alns.end(); ++a) {
    const uint offset = a->width + width - 1;
    uint danger_zone = 0;
    for (uint s = 0; s < starts.size(); ++s)
      if (a->starts[s] + offset == starts[s] && a->starts[s] != 0)
	++danger_zone;
    if (danger_zone > !args::strict_subopt)
      return true;
  }

  return false;
}

// Holds the right end of the motif fixed, and varies the left end
bool glam::resize_left()
{
  //  cerr << "resize_left" << endl;
  const float col_score = precalc_column[block_num];
  vector<float> scores;
  scores.reserve(max_poss_width + 1 - args::min_width);

  // reverse left & right (this is a bit kludgy):
  const uint to_end = align.width - 1;
  for (vector<uint>::iterator s = align.starts.begin(); s != align.starts.end(); ++s)
    if (*s != 0)
      *s += to_end;
  reverse(count_matrix.begin(), count_matrix.end());

  count_matrix.resize(args::min_width);
  float score = 0;
  scores.push_back(score);

  // consider block_num == 0
  for (uint k = args::min_width; k < max_poss_width; ++k) {
    if (resize_left_clash(old_alns, align.starts, k))
      break;
    vector<uint> counts(alphsize, 0);
    if (!count_column(counts, align.starts, -k))  // stop if hit a masked base
      break;
    count_matrix.push_back(counts);
    score += inv_temp * (col_score + counts_to_score(counts));
    scores.push_back(score);
  }

  const uint old_width = count_matrix.size();
  align.width = choose_score(scores) + args::min_width;
  count_matrix.resize(align.width);

  // re-reverse left & right:
  const uint to_beg = align.width - 1;
  for (vector<uint>::iterator s = align.starts.begin(); s != align.starts.end(); ++s)
    if (*s != 0)
      *s -= to_beg;
  reverse(count_matrix.begin(), count_matrix.end());

  return align.width != old_width;
}

// Perform one alignment refinement step (or attempt to)
bool glam::do_move()
{
  const static float resize_prob
    = 2 * args::resize_freq / (seqs.size() + 2 * args::resize_freq);

  if (rand(1.0f) < resize_prob)
    if (rand(2u) == 0)
      return resize_right();
    else
      return resize_left();
  else
    return gibbs_move();
}

// Get counts of ACGT in each column of an alignment
// assumes masked bases won't be encountered
void glam::get_count_matrix(vector<vector<uint> > & count_mat, const alignment & aln)
{
  count_mat.assign(aln.width, vector<uint>(alphsize, 0u));

  for (uint s = 0; s < aln.starts.size(); ++s)
    if (aln.starts[s] != 0)  // if this sequence is in the alignment
      count_seq(count_mat, seqs[s].begin() + aln.starts[s]);
}

// Return true if the alignments have any aligned pair of bases in common
bool glam::aln_overlap(const alignment & a1, const alignment & a2)
{
  assert(a1.starts.size() == a2.starts.size());
  vector<int> offsets;

  for (uint s = 0; s != a1.starts.size(); ++s)
    if (ranges_overlap(a1.starts[s], a1.width, a2.starts[s], a2.width))
      offsets.push_back(int(a2.starts[s]) - int(a1.starts[s]));

  sort(offsets.begin(), offsets.end());
  return adjacent_find(offsets.begin(), offsets.end()) != offsets.end();
}

// Does alignment a overlap any of the alignments in v?
bool glam::any_aln_overlap(const alignment & a, const vector<alignment> & v)
{
  for (vector<alignment>::const_iterator i = v.begin(); i != v.end(); ++i)
    if (aln_overlap(a, *i))
      return true;

  return false;
}

// Return true if the sequence segment overlaps a masked base
bool glam::overlaps_mask(const vector<uint> & seq, const uint start, const uint width)
{
  if (start > 0)  // the block is present in this sequence
    for (uint k = 0; k < width; ++k)
      if (seq[start + k] >= alphsize)
	return true;

  return false;
}

// Return true if the sequence segment overlaps any previous alignment
bool glam::overlaps_alns(const vector<alignment> & alns, const uint start, const uint width, const uint seq)
{
  for (vector<alignment>::const_iterator a = alns.begin();
       a != alns.end(); ++a)
    if (ranges_overlap(start, width, a->starts[seq], a->width))
      return true;

  return false;
}

// Get a random alignment, avoiding masked positions,
// and avoiding previous alignments if in 'hard-exclude' mode
void glam::random_aln(alignment & a)
{
  a.starts.clear();
  a.width = random_choice(width_probs.begin(), width_probs.end(), 1.0f)
    - width_probs.begin() + args::min_width;

  for (uint s = 0; s < seqs.size(); ++s) {
    uint start;
    do {
      start = rand((uint)seqs[s].size() - 1 - !args::zoops) + !args::zoops;
    } while (overlaps_mask(seqs[s], start, a.width) ||
	     args::strict_subopt && overlaps_alns(old_alns, start, a.width, s));
    a.starts.push_back(start);
  }
}

// Get a random starting alignment, avoiding overlap with previous alignments
void glam::random_start()
{
  const uint max_trials = 100000;

  uint trial;
  for (trial = 0; trial < max_trials; ++trial) {
    random_aln(align);
    if (!any_aln_overlap(align, old_alns))
      break;
  }

  if (trial == max_trials)
    mcf::die("Couldn't find any alignment that doesn't overlap a previous alignment after " + mcf::tostring(max_trials) + " trials");

  block_num = align.starts.size()
    - count(align.starts.begin(), align.starts.end(), 0u);
  get_count_matrix(count_matrix, align);
}

// Get target accept rate for modified lam schedule
// depends on how far we've got through total # of iterations
float target_accept_rate(const float time_fraction)
{
  if (time_fraction < 0.15f)  // initial geometric decay:
    return 0.44f + 0.56f * pow(560.0f, -time_fraction / 0.15f);
  else if (time_fraction < 0.65f)
    return 0.44f;  // magic 44%
  else  // final geometric decay:
    return 0.44f * pow(440.0f, -(time_fraction - 0.65f) / 0.35f);
}

// Do a sampling run, from a random start
void glam::anneal(result_struct & best)
{
  const static float cool = 1.0 / args::cool_factor;
  // ewma = exponentially weighted moving average
  const float ewma_param = 499.0 / 500.0;

  random_start();
  inv_temp = 1.0 / args::init_temp;  // breaks if init_temp = 0!
  float accept_rate = 0.5;
  uint no_improvement = 0;  // count iterations without improvement
  best.aln = align;
  best.block_num = block_num;
  best.score = get_score(count_matrix, block_num);

  uint t;
  for (t = 0;
       args::lam && t < args::stop_after || !args::lam && no_improvement < args::stop_after;
       ++t) {
    if (inv_temp > 1e+10)  // ensure minus_infinity*inv_temp doesn't overflow
      inv_temp = 1e+10;
    const bool accepted = do_move();
    const float score = get_score(count_matrix, block_num);
    accept_rate = ewma_param * accept_rate + (1.0f - ewma_param) * accepted;
    if (args::profile)  // -score = "energy"
      cout << -score << '\t' << accept_rate << '\t' << 1.0f / inv_temp << '\t'
	   << count_matrix.size() << '\t' << block_num << '\n';
    if (args::lam) {
      float target = target_accept_rate(float(t) / args::stop_after);
      if (accept_rate > target)
	inv_temp *= cool;
      else
	inv_temp /= cool;
    } else
      inv_temp *= cool;
    if (score > best.score) {
      best.aln = align;
      best.block_num = block_num;
      best.score = score;
      no_improvement = 0;
    } else
      ++no_improvement;
  }
  cerr << t << " iterations" << endl;
}

/****** Initialization & preprocessing ******/

// Mask bases of the sequence that participate in any alignment
void glam::mask_alns(vector<uint> & seq, const vector<alignment> & alns, const uint s)
{
  for (uint a = 0; a < alns.size(); ++a)
    if (alns[a].starts[s] != 0)
      for (uint k = 0; k < alns[a].width; ++k)
	seq[alns[a].starts[s] + k] = alphsize;
}

// Add # of ways blocks of different widths can fit in a run of unmasked bases
void glam::add_width_ways(vector<uint> & num_ways, const uint run)
{
  if (num_ways.size() < run)
    num_ways.resize(run);
  for (uint w = 0; w < run; ++w)  // w=0 --> width=1, etc
    num_ways[w] += run - w;
}

// Get a priori probabilities of different alignment widths
// (doing this is probably beyond the call of duty!)
// account for masked bases, and previous alignments if in 'hard-exclude' mode
// assume probs is empty
unsigned glam::get_width_probs(vector<float> & probs)
{
  for (uint s = 0; s < seqs.size(); ++s) {
    vector<uint> seq(seqs[s]);
    if (args::strict_subopt)
      mask_alns(seq, old_alns, s);
    vector<uint> num_ways;  // # of ways a block of width w can fit in a seq
    uint run = 0;  // counts runs of non-mask characters

    for (vector<uint>::const_iterator n = seq.begin(); n != seq.end(); ++n)
      if (*n < alphsize)
        ++run;
      else {
	add_width_ways(num_ways, run);
	run = 0;
      }
    add_width_ways(num_ways, run);  // don't assume sequence ends with mask

    if (s == 0 ||
	args::zoops == probs.size() < num_ways.size())
      probs.resize(num_ways.size());

    for (uint w = 0; w < min(num_ways.size(), probs.size()); ++w)
      // work in log space to avoid overflow:
      probs[w] += log(float(num_ways[w] + args::zoops));  // +1 if ZOOPS
  }

  if (probs.size() < args::min_width)
    mcf::die("Error: max possible alignment width = " + mcf::tostring(probs.size()) + ", but min width = " + mcf::tostring(args::min_width));

  if (probs.size() > args::max_width)
    probs.resize(args::max_width);  // assume max_width >= min_width
  probs.erase(probs.begin(), probs.begin() + args::min_width - 1);
  rescale_exp(probs);
  mcf::normalize(probs.begin(), probs.end());
  return probs.size() + args::min_width - 1;
}

// Stick mask characters on to beginnings and ends of sequences
// then stick reverse complements on to ends of sequences
void glam::preprocess_seqs(const bool one_strand)
{
  for (uint s = 0; s < seqs.size(); ++s) {
    seqs[s].insert(seqs[s].begin(), alphsize);  // stick mask char at start
    seqs[s].push_back(alphsize);  // stick mask character at end
    if (!one_strand)
      // stick reverse complement on to end of sequence
      // including initial mask character, which is duplicated to the very end
      for (int n = seqs[s].size() - 2; n >= 0; --n) {
	const uint base = seqs[s][n];
	seqs[s].push_back(base < alphsize ? alphsize - base - 1 : alphsize);
      }
  }
}

// Initialize the PSSM used in normal Gibbs moves
// memory is cheap, don't worry about allocating more than might be needed
void glam::init_gibbs_mat(const uint width)
{
  //  gibbs_mat = new_matrix<float>(width, alphsize + 1);
  //  gibbs_mat = (float (*) [alphsize+1])malloc(width * sizeof(float[alphsize+1]));
  gibbs_mat = new float[width][alphsize+1];
  //  gibbs_mat = matrix<float>(width, alphsize + 1);

  // set scores for masked bases to minus_infinity:
  for (uint k = 0; k < width; ++k)
    gibbs_mat[k][alphsize] = minus_infinity;
}

// Initialization that needs to be performed once only
void glam::init()
{
  preprocess_seqs(args::one_strand);
  max_poss_width = get_width_probs(width_probs);  // do after preprocess_seqs
  init_gibbs_mat(max_poss_width);
  srand(args::random_seed);
  const uint seqnum = seqs.size();

  if (args::even_pseudos)
    pseudos.assign(alphsize, args::pseudocount_weight / alphsize);
  else
    transform(bg_probs.begin(), bg_probs.end(), back_inserter(pseudos),
	      bind2nd(multiplies<float>(), args::pseudocount_weight));

  // calculate "prior_bit" - comes from prior pdf of column probabilites:
  float prior_bit = lgamma(args::pseudocount_weight);
  for (uint i = 0; i < alphsize; ++i)
    prior_bit -= lgamma(pseudos[i]);

  // score for adding 1 column (depends on # of sequences containing a block):
  for (uint b = 0; b <= seqnum; ++b)
    precalc_column.push_back(prior_bit - lgamma(args::pseudocount_weight + b));

  // score for the block being absent in 1 sequence (depends on # of other sequences containing a block):
  for (uint b = 0; b < seqnum; ++b)
    precalc_absent.push_back(log(args::pseudocount_weight + b));

  vector<float> log_bg_probs;
  for (uint i = 0; i < alphsize; ++i)
    // assume we never get log(0), since we use a pseudocount to get bg_probs
    log_bg_probs.push_back(log(bg_probs[i]));

  precalc_gibbs_mat = matrix<float>(seqnum, alphsize);
  for (uint c = 0; c < seqnum; ++c)
    for (uint i = 0; i < alphsize; ++i)
      precalc_gibbs_mat[c][i] = log(c + pseudos[i]) - log_bg_probs[i];

  precalc_resize_scores = matrix<float>(seqnum + 1, alphsize);
  for (uint c = 0; c <= seqnum; ++c)
    for (uint i = 0; i < alphsize; ++i)
      precalc_resize_scores[c][i] = lgamma(c + pseudos[i]) - c * log_bg_probs[i];
}

// Read sequences from the file, and calculate probs of A, C, G, T
void glam::read_seqs(const string & filename, const bool one_strand)
{
  ifstream file(filename.c_str());
  if (!file) mcf::die("Sorry, couldn't open file " + filename);
  unsigned (*translator)(char) = (args::mask_lower ? nolower_translator : mcf::DNA_to_number);

  vector<uint> base_counts;
  vector<uint> seq;
  string name;
  while(mcf::get_fasta(file, seq, name, translator)) {
    mcf::count_residues(seq, base_counts, alphsize);
    seq_lengths.push_back(seq.size());
    seqs.push_back(vector<uint>());  // this crappy swap kludge is faster
    seq.swap(seqs.back());
    seq_names.push_back(string());
    name.swap(seq_names.back());
  }

  if (!one_strand)  // if both strands are to be examined:
    // sum counts of a and t, c and g
    for (uint i = 0; i < alphsize / 2; ++i) {
      base_counts[i] += base_counts[alphsize - i - 1];
      base_counts[alphsize - i - 1] = base_counts[i];
    }

  if (args::even_bg)
    bg_probs.assign(alphsize, 1.0 / alphsize);
  else {
    const float bg_pseudo = 1;  // pseudocount
    transform(base_counts.begin(), base_counts.end(), back_inserter(bg_probs),
	      bind2nd(plus<uint>(), bg_pseudo));
    mcf::normalize(bg_probs.begin(), bg_probs.end());
  }
}

void glam::print_misc_info(ostream & strm)
{
  strm
    << seqs.size() << " sequences in file\n"
    << "Residue abundances:"
    << " a=" << bg_probs[0] << " c=" << bg_probs[1]
    << " g=" << bg_probs[2] << " t=" << bg_probs[3] << '\n'
    << "Pseudocounts:"
    << " a=" << pseudos[0] << " c=" << pseudos[1]
    << " g=" << pseudos[2] << " t=" << pseudos[3] << '\n'
    << "Max possible alignment width: " << max_poss_width << '\n'
    << flush;
}

// Is the alignment in the forward strand of the topmost sequence?
bool glam::canonical_orientation(const alignment & a, const vector<uint> & seqlens)
{
  assert(seqlens.size() == a.starts.size());

  for (uint i = 0; i != seqlens.size(); ++i)
    if (a.starts[i] > seqlens[i])
      return false;
    else if (a.starts[i] != 0)
      return true;

  return true;  // degenerate case: alignment absent in every sequence
}

// Flip an alignment onto the opposite strands
void glam::reflect_alignment(alignment & a, const vector<uint> & seqlens)
{
  assert(seqlens.size() == a.starts.size());

  for (uint i = 0; i != seqlens.size(); ++i)
    if (a.starts[i] != 0)
      a.starts[i] = 2 * seqlens[i] + 3 - a.starts[i] - a.width;
}

// Add a new result to the results vector
// if we've seen an identical result before, increment this result's count
void glam::add_to_results(result_struct & res)
{
  if (!canonical_orientation(res.aln, seq_lengths))
    reflect_alignment(res.aln, seq_lengths);

  for (uint r = 0; r < results.size(); ++r)
    if (res.aln == results[r].aln) {
      ++results[r].count;
      return;
    }

  res.count = 1;
  results.push_back(res);
}

// Get the first word in string s, and truncate it at maxlen characters
std::string first_word(const std::string & s, const unsigned maxlen)
{
  std::string w;
  std::istringstream bogus(s);
  bogus >> w;
  return w.substr(0, maxlen);
}

// Return the marginal score for a sequence in an alignment
float glam::get_marginal_score(const result_struct & r, const uint s)
{
  const uint pos = r.aln.starts[s];
  if (pos == 0)  // if this sequence is absent from the alignment
    return 0;
  vector<vector<uint> > count_mat;
  get_count_matrix(count_mat, r.aln);
  uncount_seq(count_mat, seqs[s].begin() + pos);
  inv_temp = 1;  // global variable used by construct_gibbs_matrix
  construct_gibbs_matrix(count_mat, r.block_num - 1);

  float score = 0;
  for (uint k = 0; k < r.aln.width; ++k)
    score += gibbs_mat[k][seqs[s][pos+k]];

  return score;
}

// This is surprisingly painful
void glam::print_alignment(const result_struct & r)
{
  const static uint coord_field
    = num_digits(*max_element(seq_lengths.begin(), seq_lengths.end()));
  const uint name_field = 12;
  const alignment & a = r.aln;
  const uint abs_field = (coord_field + 1 + args::flank_seq) * 2 + a.width + 3;

  for (uint s = 0; s < seqs.size(); ++s) {
    cout.setf(ios::left, ios::adjustfield);
    cout << setw(name_field) << first_word(seq_names[s], name_field).c_str() << ' ';
    cout.setf(ios::right, ios::adjustfield);

    if (a.starts[s] == 0) {  // sequence isn't in alignment
      cout.setf(ios::left, ios::adjustfield);
      cout << setw(abs_field) << "absent";
      cout.setf(ios::right, ios::adjustfield);

    } else if (a.starts[s] <= seq_lengths[s]) {  // on + strand
      const uint start = a.starts[s];
      const uint left_flank = min(args::flank_seq, start - 1);
      const uint right_flank = min(args::flank_seq, seq_lengths[s] + 1 - start - a.width);

      cout << setw(coord_field) << start - left_flank << ' ';
      for (uint i = 0; i < args::flank_seq - left_flank; ++i)
	cout << ' ';
      for (uint i = left_flank; i > 0; --i)
	cout << mcf::number_to_DNA(seqs[s][start - i]);
      for (uint i = 0; i < a.width; ++i)
	cout << char(toupper(mcf::number_to_DNA(seqs[s][start + i])));
      for (uint i = 0; i < right_flank; ++i)
	cout << mcf::number_to_DNA(seqs[s][start + a.width + i]);
      for (uint i = 0; i < args::flank_seq - right_flank; ++i)
	cout << ' ';
      cout << ' ' << setw(coord_field) << start + a.width + right_flank - 1;
      cout << "  +";

    } else {  // on - strand
      const uint start = seqs[s].size() - a.starts[s] - 1;
      const uint left_flank = min(args::flank_seq, seq_lengths[s] - start);
      const uint right_flank = min(args::flank_seq, start - a.width);

      cout << setw(coord_field) << start + left_flank << ' ';
      for (uint i = 0; i < args::flank_seq - left_flank; ++i)
	cout << ' ';
      for (uint i = left_flank; i > 0; --i)
	cout << mcf::number_to_DNA(3 - seqs[s][start + i]);
      for (uint i = 0; i < a.width; ++i)
	cout << char(toupper(mcf::number_to_DNA(3 - seqs[s][start - i])));
      for (uint i = 0; i < right_flank; ++i)
	cout << mcf::number_to_DNA(3 - seqs[s][start - a.width - i]);
      for (uint i = 0; i < args::flank_seq - right_flank; ++i)
	cout << ' ';
      cout << ' ' << setw(coord_field) << 1 + start - a.width - right_flank;
      cout << "  -";

    }

    cout << "  (" << get_marginal_score(r, s) << ")" << endl;
  }

  cout << setw(name_field + coord_field + 2 + args::flank_seq) << "";
  for (uint i = 0; i < a.width; ++i) {
    vector<uint> counts(alphsize, 0);
    count_column(counts, a.starts, i);
    float score = precalc_column[r.block_num] + counts_to_score(counts);
    cout << (score > 0 ? '*' : ' ');
  }
  cout << endl;
}

// Used to sort results in reverse order of score
bool by_score(const glam::result_struct & a, const glam::result_struct & b)
{  return b.score < a.score;  }

void glam::print_results(const bool one_strand, float k, float h)
{
  sort(results.begin(), results.end(), by_score);

  if (args::want_evalue)
    k = one_strand ? log(k) : log(k) + (seqs.size() - 1) * log(2.0f);

  for (uint r = 0; r < results.size(); ++r) {
    if (r == 0)
      cout << "Best alignment found:\n";
    else if (r == 1)
      cout << "Other alignments:\n";
    cout.precision(6);  // higher precision for alignment scores
    cout
      << "Score: " << results[r].score / log(2.0f) << " bits"
      << "  Width: " << results[r].aln.width
      << "  Sequences: " << results[r].block_num
      << "  Runs: " << results[r].count
      ;
    cout.precision(3);  // 3 sig figs for evalues & marginal scores
    if (args::want_evalue)
      cout << "  E-value: " << stats::evalue(results[r].score, k, h, seq_lengths);
    cout << endl;
    if (r == 0 || args::all_aligns) {
      print_alignment(results[r]);
      cout << endl;
    }
  }
  if (results.size() > 1 && !args::all_aligns)
    cout << endl;
}

// I'll probably move this to a library
inline std::ifstream & open_or_die(const std::string & filename, std::ifstream & strm)
{
  strm.open(filename.c_str());
  if (!strm) mcf::die("Sorry, couldn't open file " + filename);
  return strm;
}

// Read all the contents of a file into a string or die
void slurp_or_die(const std::string & filename, std::string & str)
{
  std::ifstream file;
  open_or_die(filename, file);
  char c;
  while (file.get(c))
    str += c;  // might not be the most efficient way
  if (!file.eof() || file.bad())
     mcf::die("Sorry, an error occurred while reading file " + filename);
}

int main(int argc, char **argv)
{
  using namespace std;
  cout << "GLAM: Gapless Local Alignment of Multiple sequences\n"
    "Compiled on " __DATE__ "\n" << endl;

  args::parse(argc, argv);
  glam::read_seqs(args::seqfile, args::one_strand);

  string old_results;
  if (!args::old_glam_file.empty()) {
    slurp_or_die(args::old_glam_file, old_results);
    glam::parse_glam(old_results, glam::old_alns, glam::seq_lengths);
    if (glam::old_alns.empty())
      cerr << "Warning: no alignments read from " << args::old_glam_file
	   << "\n";
  }

  glam::init();
  cout.precision(3);  // for printing energies, etc in profile mode

  for (unsigned r = 0; r < args::runs; ++r) {
    cerr << "Run " << r+1 << "... " << flush;
    if (args::profile) {
      cerr << endl;
      cout << "#Energy\tAccept\tTemp\tWidth\tSeqs\n";
    }
    glam::result_struct res;
    glam::anneal(res);
    glam::add_to_results(res);
    if (args::profile)
      cout << endl;
  }

  float k, h;  // parameters for calculating evalues
  if (args::want_evalue)
    stats::params(k, h, glam::bg_probs, glam::pseudos, glam::seqs.size());

  cerr << endl;
  glam::print_results(args::one_strand, k, h);
  cout.precision(6);  // reset to default precision
  cout << args::command << endl;
  glam::print_misc_info(cout);
  if (args::want_evalue)
    cout << "K: " << k << "\nH: " << h << endl;
  if (!old_results.empty())
    cout << "\n*** Previous results ***\n" << old_results;
}
