#include <nksprobdist.h>

Nks_FiniteIntProbDist::Nks_FiniteIntProbDist (// contains parameters of a finite span probability distribution 
size_t dimProb_,                     // #(distinct values)          
long *value_,                        // values 
double *prob_) :                     // probability of corresponding value 
// Assumptions:
//    (Error) value_ is in increasing order 
//    (Error) prob_ elements are positive 
dimProb (dimProb_),
value (0),
prob (0),
mu (0.0),
sigma (0.0)
{
MSG("Entering Nks_FiniteIntProbDist::Nks_FiniteIntProbDist 3")
   if (dimProb_ == 0) {
return;
   }
   for (size_t i = 0; i < dimProb_; i++) {
      if (1 <= i && value_ [i] <= value_ [i - 1]) {
Nks_AbortFormat ("Nks_FiniteIntProbDist::Nks_FiniteIntProbDist 3 : "
                 "(Error) value_ [%ld] <= value_ [%ld]", (long) i, (long) (i - 1));
      } else if (prob_ [i] < 0.0) {
Nks_Abort ("Nks_FiniteIntProbDist::Nks_FiniteIntProbDist 3 : (Error) prob_ [i] < 0.0");
      } else if (1.0 < prob_ [i]) {
Nks_Abort ("Nks_FiniteIntProbDist::Nks_FiniteIntProbDist 3 : (Error) 1.0 < prob_ [i]");
      }
   }
   NKS_CATCH(value = new long [dimProb_],  // values 
      std::bad_alloc, "Nks_FiniteIntProbDist::Nks_FiniteIntProbDist 3");
   Nks_MemCopy ((void *) value, (void *) value_, dimProb_ * sizeof (long));
   NKS_CATCH(prob = new double [dimProb_], // probability of corresponding value 
      std::bad_alloc, "Nks_FiniteIntProbDist::Nks_FiniteIntProbDist 3");
   Nks_MemCopy ((void *) prob, (void *) prob_, dimProb_ * sizeof (double));
   mu = CalcMu ();                   // step mean
   sigma = CalcSigma ();             // step standard deviation
MSG("Leaving Nks_FiniteIntProbDist::Nks_FiniteIntProbDist 3")
}

Nks_FiniteIntProbDist::Nks_FiniteIntProbDist (// contains parameters of an integer probability distribution  
Nks_LongMeasure *longMeasure_,       // function returning a probability measure on the integers          
void *param_,                        // parameters for longMeasure_
long lower_,                         // lower bound of explicitly calculated values
long upper_) :                       // upper bound of explicitly calculated values
dimProb (0),
value (0),
prob (0),
mu (0.0),
sigma (0.0)
{
MSG("Entering Nks_FiniteIntProbDist::Nks_FiniteIntProbDist 4")
   if (longMeasure_ == 0 ||          // no function given
       upper_ < lower_) {            // no probability values stored
return;
   } else {
      dimProb = (size_t) (upper_ - lower_ + 1);
      NKS_CATCH(value = new long [dimProb],  // values 
         std::bad_alloc, "Nks_FiniteIntProbDist::Nks_FiniteIntProbDist 4");
      NKS_CATCH(prob = new double [dimProb], // probability of corresponding value 
         std::bad_alloc, "Nks_FiniteIntProbDist::Nks_FiniteIntProbDist 4");
      long val = 0;
      size_t i = 0;
      for (val = lower_, i = 0; val <= upper_; val++, i++) {
         value [i] = val;
         prob [i] = (*longMeasure_) (val, param_);
      }
      mu = CalcMu ();                   // step mean
      sigma = CalcSigma ();             // step standard deviation
   }
MSG("Leaving Nks_FiniteIntProbDist::Nks_FiniteIntProbDist 4")
}

Nks_FiniteIntProbDist::Nks_FiniteIntProbDist ( // contains parameters of a finite span probability distribution 
const Nks_FiniteIntProbDist &finiteIntProbDist_) // finite span probability distribution
{
MSG("Entering Nks_FiniteIntProbDist::Nks_FiniteIntProbDist 1")
   Nks_Initialize (*this, finiteIntProbDist_);
MSG("Leaving Nks_FiniteIntProbDist::Nks_FiniteIntProbDist 1")
}

Nks_FiniteIntProbDist &Nks_FiniteIntProbDist::operator= (// contains parameters of a finite span probability distribution 
const Nks_FiniteIntProbDist &finiteIntProbDist_) // finite span probability distribution
{
MSG("Entering Nks_FiniteIntProbDist::Nks_FiniteIntProbDist 1")
// protect against self-assignment
   if (this  == &finiteIntProbDist_) {
return *this;
   }
// delete old values
   this->~Nks_FiniteIntProbDist ();
// initialize
   Nks_Initialize (*this, finiteIntProbDist_);
return *this;
MSG("Leaving Nks_FiniteIntProbDist::Nks_FiniteIntProbDist 1")
}

Nks_FiniteIntProbDist::~Nks_FiniteIntProbDist ()// contains parameters of a finite span probability distribution 
{
MSG("Entering Nks_FiniteIntProbDist::~Nks_FiniteIntProbDist")
   delete [] value;
   value = 0;
   delete [] prob;
   prob = 0;
MSG("Leaving Nks_FiniteIntProbDist::~Nks_FiniteIntProbDist")
}

double Nks_FiniteIntProbDist::CalcMu ()// calculates mean  
{
MSG("Entering Nks_FiniteIntProbDist::CalcMu")
   if (dimProb == 0) {
return 0.0;
   }
   double sum = 0.0;
   for (size_t i = 0; i < dimProb; i++) {
      sum += prob [i] * (double) value [i];
   }
MSG("Leaving Nks_FiniteIntProbDist::CalcMu")
return sum;
}

double Nks_FiniteIntProbDist::CalcSigma ()// calculates standard deviation 
{
MSG("Entering Nks_FiniteIntProbDist::CalcSigma")
   if (dimProb == 0) {
return 0.0;
   }
   double sum = 0.0;
   for (size_t i = 0; i < dimProb; i++) {
      sum += prob [i] * (double) value [i] * (double) value [i];
   }
MSG("Leaving Nks_FiniteIntProbDist::CalcSigma")
return NKS_SQRT (sum - mu * mu);
}

void Nks_FiniteIntProbDist::Nks_Initialize ( // contains parameters of a finite span probability distribution 
Nks_FiniteIntProbDist &initial,          // alias to initialization
const Nks_FiniteIntProbDist &finiteIntProbDist_) // finite span probability distribution
{
MSG("Entering Nks_FiniteIntProbDist::Nks_Initialize")
   initial.dimProb = finiteIntProbDist_.dimProb;
// exceptional value
   if (finiteIntProbDist_.dimProb == 0) {
      initial.value = 0;
      initial.prob = 0;
      initial.mu = 0.0;
      initial.sigma = 0.0;
return;
   }
// allocate
   NKS_CATCH(initial.value = new long [finiteIntProbDist_.dimProb],  // values 
      std::bad_alloc, "Nks_FiniteIntProbDist::Nks_Initialize");
   NKS_CATCH(initial.prob = new double [finiteIntProbDist_.dimProb], // probability of corresponding value 
      std::bad_alloc, "Nks_FiniteIntProbDist::Nks_Initialize");
// copy in new elements
   Nks_MemCopy ((void *) initial.value, (void *) finiteIntProbDist_.value, finiteIntProbDist_.dimProb * sizeof (long));
   Nks_MemCopy ((void *) initial.prob, (void *) finiteIntProbDist_.prob, finiteIntProbDist_.dimProb * sizeof (double));
   initial.mu = CalcMu ();                   // step mean
   initial.sigma = CalcSigma ();             // step standard deviation
MSG("Leaving Nks_FiniteIntProbDist::Nks_Initialize")
}

Nks_InfiniteIntProbDist::Nks_InfiniteIntProbDist (// contains parameters of an integer probability distribution  
Nks_LongMeasure *longMeasure_,     // function returning a probability measure on the integers          
void *param_,                      // parameters for longMeasure_
long lower_,                       // lower bound of explicitly calculated values
long upper_) :                     // upper bound of explicitly calculated values
Nks_FiniteIntProbDist (longMeasure_, param_, lower_, upper_)
{
MSG("Entering Nks_InfiniteIntProbDist::Nks_InfiniteIntProbDist 4")
   longMeasure = longMeasure_;       // function returning a probability measure on the integers          
   if (longMeasure_ == 0) {
      longMeasure = 0;
      param = 0;
      lower = LOWER_DUMMY;
      upper = UPPER_DUMMY;
return;
   }
   param = param_;                   // parameters for longMeasure
   if (upper_ < lower_) {            // no probability values stored
      lower = LOWER_DUMMY;
      upper = UPPER_DUMMY;
   } else {
      lower = lower_;
      upper = upper_;
   }
   probabilitySurplus = 1.0;
   for (size_t i = 0; i < dimProb; i++) { // maintains correctness for dimProb = 0
      probabilitySurplus -= prob [i];
   }
MSG("Leaving Nks_InfiniteIntProbDist::Nks_InfiniteIntProbDist 4")
}

Nks_InfiniteIntProbDist::Nks_InfiniteIntProbDist ( // contains parameters of an integer probability distribution 
const Nks_InfiniteIntProbDist &infiniteIntProbDist_) : // integer probability distribution
// initialize the immediate base
Nks_FiniteIntProbDist ((const Nks_FiniteIntProbDist &) infiniteIntProbDist_)
{
MSG("Entering Nks_InfiniteIntProbDist::Nks_InfiniteIntProbDist 1")
// initialize the members
   Nks_Initialize (*this, infiniteIntProbDist_);
MSG("Leaving Nks_InfiniteIntProbDist::Nks_InfiniteIntProbDist 1")
}

Nks_InfiniteIntProbDist &Nks_InfiniteIntProbDist::operator= (// contains parameters of a finite span probability distribution 
const Nks_InfiniteIntProbDist &infiniteIntProbDist_) // finite span probability distribution
{
MSG("Entering Nks_InfiniteIntProbDist::Nks_InfiniteIntProbDist 1")
// protect against self-assignment
   if (this == &infiniteIntProbDist_) {
return *this;
   }
// initialize the base classes
   *(Nks_FiniteIntProbDist *) this = (const Nks_FiniteIntProbDist) infiniteIntProbDist_;
// initialize the novel class members
   Nks_Initialize (*this, infiniteIntProbDist_);
return *this;
MSG("Leaving Nks_InfiniteIntProbDist::Nks_InfiniteIntProbDist 1")
}

Nks_InfiniteIntProbDist::~Nks_InfiniteIntProbDist () // contains parameters of an integer probability distribution 
{
MSG("Entering Nks_InfiniteIntProbDist::~Nks_InfiniteIntProbDist")
MSG("Leaving Nks_InfiniteIntProbDist::~Nks_InfiniteIntProbDist")
}

void Nks_InfiniteIntProbDist::ChangeLowerUpper ( // change the stored arrays
long lower_,                            // new value of lower
long upper_)                            // new value of upper
{
// delete arrays
   ((Nks_FiniteIntProbDist *) this)->~Nks_FiniteIntProbDist ();
// adjust members accordingly
   *this = Nks_InfiniteIntProbDist (longMeasure, param, lower_, upper_);
}

void Nks_InfiniteIntProbDist::Nks_Initialize ( // contains parameters of a finite span probability distribution 
Nks_InfiniteIntProbDist &init,               // alias to initialization
const Nks_InfiniteIntProbDist &infiniteIntProbDist_)  // finite span probability distribution
{
MSG("Entering Nks_FiniteIntProbDist::Nks_Initialize")
   (Nks_FiniteIntProbDist) init = (Nks_FiniteIntProbDist) infiniteIntProbDist_;
   init.longMeasure = infiniteIntProbDist_.longMeasure; // function returning a probability measure on the integers         
   init.param = infiniteIntProbDist_.param;   // parameters for longMeasure
   init.lower = infiniteIntProbDist_.lower;   // lower bound of explicitly calculated values
   init.upper = infiniteIntProbDist_.upper;   // upper bound of explicitly calculated values
   init.probabilitySurplus = infiniteIntProbDist_.probabilitySurplus; // extra probability outside of range
MSG("Leaving Nks_FiniteIntProbDist::Nks_Initialize")
}
