#include <nksrootfinder.h>

const size_t ITMAX = 100;
#define F(x)  ((*f)(x) - y)
#define DF(x) ((*df)(x))

double Nks_RootFinder (                      // finds root f(x) = y with derivative y'
double y,                                    // ordinate
double (*f) (double x),                      // function : param contains parameters
double (*df) (double x),                     // derivative of function
double p,                                    // interval end-point
double x,                                    // initial guess 
double q,                                    // interval end-point 
double tol,                                  // absolute tolerance
double rtol)                                 // relative tolerance
/*
Bisection routine that locates x : f(x) = y to within absolute error +-tol
Uses derivative information through a Newton-Raphson alternative
Assumptions:
	(Error) f(p) <= y <= f(q) or f(q) <= y <= f(p)
*/
{
// Preliminary checks for improper bracketing and end-point root.
	if (F(p) * F(q) > 0.0)
Nks_Abort ("Nks_RootFinder : root not bracketed");
   if (F(p) == 0.0) {
return p;
   } else if (F(q) == 0.0) {
return q;
   }
// Swaps end-points if necessary to make F(p) < 0.0 < F(q)
	if (F(p) > 0.0) {
   	double temp = p;           // storage for swapping end-points 
		p = q;
		q = temp;
	}
// Set up the Bisection & Newton-Raphson iteration.
   if ((x-p) * (x-q) > 0.0) {
		x = 0.5 * (p + q);
   }
	double dx;                    // present interval length
	double dxold;                 // old interval length
	double fx;                    // f(x)-y
	double dfx;                   // Df(x)
	dxold = dx = p - q;
	for (size_t j = 1; j <= ITMAX; j++) {
		fx = F(x);    
		if (fx == 0.0) {           // Check for termination.
return x;
		} else if (fx < 0.0) {
			p = x;
		} else {
			q = x;
		}
		dfx = DF(x);
// Check: root out of bounds or bisection faster than Newton-Raphson?
		if ((dfx * (x-p) - fx) * (dfx * (x - q) - fx) >= 0.0 || 
			2.0 * NKS_ABS(fx) > NKS_ABS(dfx * dx)) {
			dx = dxold;             // Bisect
			dxold = 0.5 * (p - q);
			x = 0.5 * (p + q);
			if (NKS_ABS(dxold) <= tol) 
return x;
		} else {     
			dx = dxold;               // Newton-Raphson
			dxold = fx / dfx;
			x -= dxold;
			if (NKS_ABS(dxold) < tol || NKS_ABS(dxold) < rtol * NKS_ABS(x)) {
				if (F(x - Nks_Signum(dxold) * tol) * fx < 0.0)	
return x;
			}
		}
	}
Nks_Abort ("Nks_RootFinder : iterations > ITMAX"); 
// Make the compiler happy and return a value : this line is never reached.
return 0.0;
}

#undef F
#undef DF
#undef ITMAX

#define F(x)  ((*f)(x) - y)

double Nks_RootBisection (                   // finds root f(x) = y
double y,                                    // f(x)=y
double (*f) (double x),                      // function : param contains parameters
double p,                                    // end-point
double q,                                    // end-point
double tol,                                  // absolute tolerance
double rtol)                                 // relative tolerance
/*
Bisection routine that locates x : f(x) = y to within absolute error +-tol
Assumptions:
	(Error) f(p) <= y <= f(q) or f(q) <= y <= f(p)
*/
{
	if (F(p) * F(q) > 0.0) {
Nks_Abort ("Nks_RootBisection : root not bracketed");
	}
	if (F(p) == 0.0) {
return p;
	} else if (F(q) == 0.0) {
return q;
	}
	if (F(p) > 0.0) {
		double temp = p;
		p = q;
		q = temp;
	}
	double x;
	for (x = 0.5 * (p + q); ! Nks_AbsRelApprox (p, x, tol, rtol); ) {
      if (F(x) < 0.0) {
			p = x;
      } else {
			q = x;
      }
      x = 0.5 * (p + q);
	}
return x;
}

#undef F

#define F(x)  ((*f)(x, param) - y)
#define DF(x) ((*df)(x, param))

double Nks_RootFinder (                      // finds root f(x) = y with derivative y'
double y,                                    // ordinate
double (*f) (double x, void *param),         // function : param contains parameters
double (*df) (double x, void *param),        // derivative of function
void *param,                                 // parameters for function
double p,                                    // interval end-point
double x,                                    // initial guess 
double q,                                    // interval end-point 
double tol,                                  // absolute tolerance
double rtol)                                 // relative tolerance
/*
Bisection routine that locates x : f(x) = y to within absolute error +-tol
Uses derivative information through a Newton-Raphson alternative
Assumptions:
	(Error) f(p) <= y <= f(q) or f(q) <= y <= f(p)
*/
{
// Preliminary checks for improper bracketing and end-point root.
	if (F(p) * F(q) > 0.0)
Nks_Abort ("Nks_RootFinder : root not bracketed");
   if (F(p) == 0.0) {
return p;
   } else if (F(q) == 0.0) {
return q;
   }
// Swaps end-points if necessary to make F(p) < 0.0 < F(q)
	if (F(p) > 0.0) {
   	double temp = p;           // storage for swapping end-points 
		p = q;
		q = temp;
	}
// Set up the Bisection & Newton-Raphson iteration.
   if ((x-p) * (x-q) > 0.0) {
		x = 0.5 * (p + q);
   }
	double dx;                    // present interval length
	double dxold;                 // old interval length
	double fx;                    // f(x)-y
	double dfx;                   // Df(x)
	dxold = dx = p - q;
	for (size_t j = 1; j <= ITMAX; j++) {
		fx = F(x);    
		if (fx == 0.0) {           // Check for termination.
return x;
		} else if (fx < 0.0) {
			p = x;
		} else {
			q = x;
		}
		dfx = DF(x);
// Check: root out of bounds or bisection faster than Newton-Raphson?
		if ((dfx * (x-p) - fx) * (dfx * (x - q) - fx) >= 0.0 || 
			2.0 * NKS_ABS(fx) > NKS_ABS(dfx * dx)) {
			dx = dxold;             // Bisect
			dxold = 0.5 * (p - q);
			x = 0.5 * (p + q);
			if (NKS_ABS(dxold) <= tol) 
return x;
		} else {     
			dx = dxold;               // Newton-Raphson
			dxold = fx / dfx;
			x -= dxold;
			if (NKS_ABS(dxold) < tol || NKS_ABS(dxold) < rtol * NKS_ABS(x)) {
				if (F(x - Nks_Signum(dxold) * tol) * fx < 0.0)	
return x;
			}
		}
	}
Nks_Abort ("Nks_RootFinder : iterations > ITMAX"); 
// Make the compiler happy and return a value : this line is never reached.
return 0.0;
}

#undef F
#undef DF
#undef ITMAX

#define F(x)  ((*f)(x, param) - y)

double Nks_RootBisection (                   // finds root f(x) = y
double y,                                    // f(x)=y
double (*f) (double x, void *param),         // function : param contains parameters
void *param,                                 // parameters for function
double p,                                    // end-point
double q,                                    // end-point
double tol,                                  // absolute tolerance
double rtol)                                 // relative tolerance
/*
Bisection routine that locates x : f(x) = y to within absolute error +-tol
Assumptions:
	(Error) f(p) <= y <= f(q) or f(q) <= y <= f(p)
*/
{
	if (F(p) * F(q) > 0.0) {
Nks_Abort ("Nks_RootBisection : root not bracketed");
	}
	if (F(p) == 0.0) {
return p;
	} else if (F(q) == 0.0) {
return q;
	}
	if (F(p) > 0.0) {
		double temp = p;
		p = q;
		q = temp;
	}
	double x;
	for (x = 0.5 * (p + q); ! Nks_AbsRelApprox (p, x, tol, rtol); ) {
      if (F(x) < 0.0) {
			p = x;
      } else {
			q = x;
      }
      x = 0.5 * (p + q);
	}
return x;
}

#undef F
