#include <nksutil.h>

/*******************************************************************************

****************************** UTILILY ROUTINES ********************************

*******************************************************************************/

void *Nks_MemNew (size_t size)
{ 
return size == (size_t) 0 ? reinterpret_cast<void *> (0) : (void *) new char [size];
}

void *Nks_MemMore (void *p, size_t size)
{
return p == reinterpret_cast<void *> (0) ? (void *) new char [size] : realloc (p, size);
}

void Nks_MemCopy (void *to, const void *from, size_t size)
{ 
	if (size != 0) {
		memcpy (to, from, size);
	}
}

void *Nks_MemFree (void *p) 
{
	delete p;
return reinterpret_cast<void *> (0);
}

void *Nks_MemSet (void *p, int c, size_t size)
{
return memset (p, c, size);
}

void *Nks_MemZero (void *p, size_t size)
{
return memset (p, '\0', size);
}

void Nks_ModularMemCopy (                          // circular copying routine
void *to,                                          // #(copied *to)
size_t toSize,                                     // #(size of *to)
size_t toMinIndex,                                 // #(placement of first copied element in *to)
void *from,                                        // #(copied *from)
size_t fromSize,                                   // #(size of *from)
size_t fromMinIndex,                               // #(placement of first copied element in *from)
size_t size,                                       // #(elements to be copied)
size_t sizeSpace)                                  // #(size of elements to be copied)
//
// Assumes : size <= MIN(fromSize, toSize)
//           fromMinIndex < fromSize
//           toMinIndex < toSize
//
{
	if (size != 0) {
      size_t copySize = NKS_MIN (size, NKS_MIN(toSize - toMinIndex, fromSize - fromMinIndex));
      memcpy ((char *) to + sizeSpace * toMinIndex, 
         (char *) from + sizeSpace * fromMinIndex, sizeSpace * copySize);
      size -= copySize;                             // amount remaining to copy
      if (size == 0) {                              // copying complete
return;
      }
      if (copySize == toSize - toMinIndex) {        // copying terminated by the end of *to
         if (copySize == fromSize - fromMinIndex) { // copying terminated by the end of *from also
            memcpy ((char *) to, (char *) from, sizeSpace * size);
return;
         }
         fromMinIndex += copySize;
         copySize = NKS_MIN (size, fromSize - fromMinIndex);
         memcpy ((char *) to, (char *) from + sizeSpace * fromMinIndex, sizeSpace * copySize);
         size -= copySize;                          // amount remaining to copy
         if (size == 0) {                           // copying complete
return;
         }
         toMinIndex += copySize;
         copySize = size;
         memcpy ((char *) to + sizeSpace * toMinIndex, (char *) from, sizeSpace * copySize);
return;
      } else if (copySize == fromSize - fromMinIndex) { // copying terminated only by the end of *from
         toMinIndex += copySize;
         copySize = NKS_MIN (size, toSize - toMinIndex);
         memcpy ((char *) to + sizeSpace * toMinIndex, (char *) from, sizeSpace * copySize);
         size -= copySize;                         // amount remaining to copy
         if (size == 0) {                          // copying complete
return;
         }
         fromMinIndex += copySize;
         copySize = size;
         memcpy ((char *) to, (char *) from + sizeSpace * fromMinIndex, sizeSpace * copySize);
return;
      }
	}
}

void *Nks::NewVector (long iMin, long iMax, size_t sizeElement)
{
   if (iMax < iMin) {
Nks_AbortFormat ("iMax = %ld < %ld = iMin\n", (long) iMax, (long) iMin);
   }
return (new char [sizeElement * (size_t) (iMax - iMin + 1)]) - sizeElement * iMin;
}

void **Nks_NewMatrix (long iMin, long iMax, long jMin, long jMax, size_t sizeElement)
{
   if (iMax < iMin) {
Nks_AbortFormat ("iMax = %ld < %ld = iMin\n", (long) iMax, (long) iMin);
   }
   if (jMax < jMin) {
Nks_AbortFormat ("jMax = %ld < %ld = jMin\n", (long) jMax, (long) jMin);
   }
   void **matrix = (new void * [(size_t) (iMax - iMin + 1)]) - iMin;
   for (size_t index = iMin; index <= iMax; index++) {
      matrix [index] = (new char [sizeElement * (size_t) (jMax - jMin + 1)]) - sizeElement * jMin;
   }
return matrix;
}

void Nks_DeleteMatrix (void **matrix, long iMin, long iMax, long jMin, long jMax, size_t sizeElement) 
// Assumes matrix was created with the call 
//         matrix = Nks_NewMatrix (iMin, iMax, jMin, jMax, sizeElement);
{
   for (long index = iMin; index <= iMax; index++) {
      delete [] ((char *) matrix [index] + sizeElement * jMin);
   }
   delete [] (matrix + iMin);
}

long **Nks_NewLongMatrix (long iMin, long iMax, long jMin, long jMax)
{
return (long **) Nks_NewMatrix (iMin, iMax, jMin, jMax, sizeof (long));
}

void Nks_DeleteLongMatrix (long **matrix, long iMin, long iMax, long jMin, long jMax) 
// Assumes matrix was created with the call 
//         matrix = Nks_NewLongMatrix (iMin, iMax, jMin, jMax);
{
   Nks_DeleteMatrix ((void **) matrix, iMin, iMax, jMin, jMax, sizeof (long));
}

