#include <algorithm>  // transform
#include <numeric>  // accumulate
#include <cmath>  // log, exp, floor, lgamma (non-standard??)

#include "nksrandomwalk.h"
#include "matrix.hpp"
#include "stats.hpp"

namespace stats {
  using mcf::matrix;
  typedef vector<pair<int, float> > int_float_vec;

  const uint alphsize = 4;

  // multiply all values by this to preserve accuracy
  // multiplying by a constant should have no effect on K
  const float mult = 100;  // 100 seems to be accurate to 2sf (for 5 seqs)

  void precalc_score_bits(matrix<float> & precalc, const vector<float> & log_probs, const vector<float> & pseudos, uint seqnum);
  void precalc_prob_bits(matrix<float> & precalc, const vector<float> & probs, uint seqnum);
  void get_distribution(const matrix<float> & precalc_scores, const matrix<float> & precalc_probs, uint seqnum, int_float_vec & score_dist);
  void eliminate_duplicates(const int_float_vec & score_dist, vector<long> & scores, vector<double> & probs);
}

// redundant with code in glam.cpp - sometimes "copy and paste" is best!
void stats::precalc_score_bits(matrix<float> & precalc, const vector<float> & log_probs, const vector<float> & pseudos, uint seqnum)
{
  precalc = matrix<float>(seqnum + 1u, alphsize);

  float pseudo_sum = accumulate(pseudos.begin(), pseudos.end(), 0.0f);

  // calculate const_bit:
  float const_bit = lgamma(pseudo_sum) - lgamma(seqnum + pseudo_sum);
  for (uint i = 0u; i < alphsize; ++i)
    const_bit -= lgamma(pseudos[i]);
  const_bit /= alphsize;

  for (uint c = 0u; c <= seqnum; ++c)
    for (uint i = 0u; i < alphsize; ++i)
      precalc[c][i] = lgamma(c + pseudos[i]) - c * log_probs[i] + const_bit;
}

// precalculate contributions to probability of getting some number of a's, c's, g's and t's
// probs: individual probabilities of acgt
// seqnum: total number of sequences
void stats::precalc_prob_bits(matrix<float> & precalc, const vector<float> & log_probs, uint seqnum)
{
  precalc = matrix<float>(seqnum + 1u, alphsize);

  float const_bit = lgamma(seqnum + 1u) / seqnum;

  for (uint c = 0u; c <= seqnum; ++c)
    for (uint i = 0u; i < alphsize; ++i)
      // Beware of overflow! This is a bit dodgy:
      precalc[c][i] = exp(c * log_probs[i] + c * const_bit - lgamma(c + 1u));
}

// I think the total number of loops will be (seqnum+1)(seqnum+2)(seqnum+3)/6
// With 100 sequences, this function is much faster than the random walk code
void stats::get_distribution(const matrix<float> & precalc_scores, const matrix<float> & precalc_probs, uint seqnum, int_float_vec & score_dist)
{
  for (uint a = 0u; a <= seqnum; ++a) {
    uint cgt = seqnum - a;
    float score1 = precalc_scores[a][0];
    float prob1 = precalc_probs[a][0];
    for (uint c = 0; c <= cgt; ++c) {
      uint gt = cgt - c;
      float score2 = score1 + precalc_scores[c][1];
      float prob2 = prob1 * precalc_probs[c][1];
      for (uint g = 0; g <= gt; ++g) {
	uint t = gt - g;
	float score4 = score2 + precalc_scores[g][2] + precalc_scores[t][3];
	float prob4 = prob2 * precalc_probs[g][2] * precalc_probs[t][3];
	int int_score = (int)floor(score4 * mult + 0.5);
	//	int int_score = (int)floor(score4 * mult);
	if (prob4 >= 1e-37)
	  score_dist.push_back(pair<int, float>(int_score, prob4));
      }
    }
  }
}

// copy vector of pairs into two separate vectors,
// eliminating duplicate scores (summing their probabilities) as we go
void stats::eliminate_duplicates(const int_float_vec & score_dist, vector<long> & scores, vector<double> & probs)
{
  if (score_dist.empty())  // for safety
    return;

  scores.push_back(score_dist[0].first);
  probs.push_back(score_dist[0].second);

  for (int_float_vec::const_iterator i = score_dist.begin() + 1; i != score_dist.end(); ++i)
    if (i->first == scores.back())
      probs.back() += i->second;
    else {
      scores.push_back(i->first);
      probs.push_back(i->second);
    }
}

void stats::params(float & k, float & h, const vector<float> & bg_probs, const vector<float> & pseudos, uint seqnum)
{
  int_float_vec score_dist;

  cerr << "Calculating score distribution... " << flush;

  {
    matrix<float> precalc_scores, precalc_probs;
    vector<float> log_probs;

    transform(bg_probs.begin(), bg_probs.end(), back_inserter(log_probs),
	      static_cast<double (*)(double)>(log));
	      // if I replace double with float, it doesn't compile - why???

    precalc_score_bits(precalc_scores, log_probs, pseudos, seqnum);
    precalc_prob_bits(precalc_probs, log_probs, seqnum);

    get_distribution(precalc_scores, precalc_probs, seqnum, score_dist);
  }

  // should sort by the first value of the pair, then the second value:
  sort(score_dist.begin(), score_dist.end());

  vector<long> scores;
  vector<double> probs;

  eliminate_duplicates(score_dist, scores, probs);

  cerr << '\n' << "Calculating random walk parameters..." << flush;
  Nks_FiniteIntRandomWalk walk(scores.size(), &scores[0], &probs[0]);

  k = float(walk.K_minus());
  h = float(walk.MuStar()) / mult;
  cerr << endl;
}

// kstuff is ln(K) for single-stranded alignments,
// or ln(K) + (N-1)*ln(2) for double-stranded alignments.
// h is the average score per column for an optimal alignment of random sequences
float stats::evalue(float score, float kstuff, float h, const vector<uint> & len)
{
  // Altschul + Gish's edge correction:
  float avglen = kstuff;
  for (vector<uint>::const_iterator i = len.begin(); i != len.end(); ++i)
    avglen += log(float(*i));
  avglen /= h;

  // Spouge's edge correction:
  //  float avglen = score / h;

  score -= kstuff;
  for (vector<uint>::const_iterator i = len.begin(); i != len.end(); ++i)
    score -= log(*i - avglen);

  return exp(-score);
}
